# coding=utf-8
import requests

from ...backend_mixin import BackendMixin
from .gcs_object_store import GCSObjectStore


class BackendGCSSignedUrlService(BackendMixin):
    def __init__(self, connection, config, handle_api):
        super(BackendGCSSignedUrlService, self).__init__(connection, config, handle_api)

    def get_signed_urls(self, method, object_names, content_type=None, **kwargs):
        headers = []
        for key, val in kwargs.items():
            headers.append('%s:%s' % (key, val))

        msg = {
            'method': method,
            'paths': object_names,
            'headers': headers,
        }

        if content_type:
            msg['content_type'] = content_type

        url = 'data_volumes/{volume_id}/gcs_urls'.format(volume_id=self._volume_id)

        result = self._handle_api(self._config, requests.post, url, msg)

        return result.get('urls', [])


class BackendGCSObjectStore(BackendMixin, GCSObjectStore):
    def __init__(self, connection, config, handle_api):
        super(BackendGCSObjectStore, self).__init__(connection, config, handle_api)
        self.__bucket_name = None
        self._signed_url_service = BackendGCSSignedUrlService(connection, config, handle_api)

    def __iter__(self):
        return super(BackendGCSObjectStore, self).__iter__()

    def add_objects(self, objects):
        return super(BackendGCSObjectStore, self).add_objects(self)

    @property
    def packs(self):
        return super(BackendGCSObjectStore, self).packs

    def contains_loose(self, sha):
        return super(BackendGCSObjectStore, self).contains_loose(sha)

    def contains_packed(self, sha):
        return super(BackendGCSObjectStore, self).contains_packed(sha)



# coding=utf-8
import datetime
import logging

import six
from ..bigquery_mixin import BigQueryMixin
from .base_metadata_db import BaseMetadataDB, MetadataTypeNotSupported
from flatten_json import unflatten


# noinspection SqlNoDataSourceInspection
class BigQueryMetadataDB(BaseMetadataDB, BigQueryMixin):
    STAGING_TABLE_NAME = 'staging'
    METADATA_TABLE_NAME = 'metadata'
    STAGING_INDEX_TABLE_NAME = 'staging_index'

    def __init__(self, connection, version=None):
        self.__version = version
        self.__staging_table = None
        self.__metadata_table = None
        self.__prev_table_info = None
        super(BigQueryMetadataDB, self).__init__(connection)

    def __get_staging_table_name(self):
        version = self.__version or 0
        return '%s_%s' % (self.STAGING_TABLE_NAME, version)

    def __get_staging_index_table_name(self):
        version = self.__version or 0
        return '%s_%s' % (self.STAGING_INDEX_TABLE_NAME, version)

    def _create_table(self):
        from google.cloud import bigquery

        schema = (
            bigquery.SchemaField('_sha', 'STRING'),
            bigquery.SchemaField('_commit_sha', 'STRING'),
            bigquery.SchemaField('_ts', 'TIMESTAMP'),
            bigquery.SchemaField('_hash', 'STRING'),
        )

        self.__metadata_table = self._create_specific_table(self.METADATA_TABLE_NAME, schema)
        self.__staging_table = self._create_specific_table(self.__get_staging_table_name(), self.__metadata_table.schema)

    def __has_column(self, column_name):
        for field in self.__metadata_table.schema:
            if column_name == field.name:
                return True

        return False

    @classmethod
    def __value_to_sql_type(cls, column_value):
        if isinstance(column_value, six.string_types):
            return 'STRING'

        if isinstance(column_value, six.integer_types):
            return 'INTEGER'

        if isinstance(column_value, float):
            return 'FLOAT'

        raise MetadataTypeNotSupported('UNKNOWN TYPE %s' % type(column_value))

    @classmethod
    def _patch_table_with_new_schema(cls, table, schema):
        table.patch(schema=schema)

    def _add_missing_columns(self, data_object):
        from google.cloud import bigquery

        new_columns = []
        for column_name, column_value in data_object.items():
            column_name = self.common_name_to_bq_field_name(column_name)
            if self.__has_column(column_name):
                continue

            new_columns.append(bigquery.SchemaField(column_name, self.__value_to_sql_type(column_value)))

        if len(new_columns) > 0:
            schema = self.__staging_table.schema[:]
            new_columns = sorted(new_columns, key=lambda field: field.name)
            schema.extend(new_columns)
            self._patch_table_with_new_schema(self.__staging_table, schema)
            self._patch_table_with_new_schema(self.__metadata_table, schema)

    def _add_data(self, flatten_data_list):
        rows = []

        now = datetime.datetime.utcnow()

        for flatten_data_list in flatten_data_list:
            row = []
            for field in self.__staging_table.schema:
                common_field_name = self.bq_field_name_to_common_name(field.name)
                if common_field_name == '_ts':
                    row.append(now)
                else:
                    row.append(flatten_data_list.get(common_field_name))

            rows.append(row)

        self.handle_bq_errors(self.__staging_table.insert_data(rows))

    def __truncate_staging(self):
        logging.info('truncate metadata staging')
        staging_table = self._get_table(self.__get_staging_table_name())
        staging_table.delete()

    def end_commit(self):
        logging.debug('bq end commit meta')

        self.__truncate_staging()

    def begin_commit(self, commit_sha, tree_id):
        logging.debug('bq begin commit meta %s %s', commit_sha, tree_id)

        from google.cloud import bigquery

        staging_table = self._get_table(self.__get_staging_table_name())
        index_staging_table = self._get_table(self.__get_staging_index_table_name())
        metadata_table = self._get_table(self.METADATA_TABLE_NAME)

        with self._connection.get_cursor() as bq_dataset:
            src_query = """
                #standardSQL
                SELECT _sha, @commit_sha as _commit_sha, *  EXCEPT(_sha, _commit_sha, index_name) FROM (
                  SELECT * EXCEPT(_max_sha, _max_ts, _hash) FROM {dataset_name}.{staging_table_name}
                  INNER JOIN (
                    SELECT _sha as _max_sha, MAX(_ts) as _max_ts
                    FROM {dataset_name}.{staging_table_name}
                    GROUP BY _sha
                  ) ON _sha = _max_sha
                  WHERE _max_ts = _ts
                )
              INNER JOIN (
                  select sha as _hash, name as index_name
                  from {dataset_name}.{staging_index_table_name}
                ) on _sha = index_name
            """.format(
                dataset_name=bq_dataset.name,
                staging_table_name=staging_table.name,
                staging_index_table_name=index_staging_table.name)

            src_query_parameters = (
                bigquery.ScalarQueryParameter('commit_sha', 'STRING', commit_sha),
            )

            self._copy_table_data(src_query, src_query_parameters, metadata_table)

    def _query_head_data(self, sha_list):
        from google.cloud import bigquery

        bq_client = self._connection

        with bq_client.get_cursor() as bq_dataset:
            query = """
            #standardSQL
            SELECT * EXCEPT(_max_sha, _max_ts, _ts) FROM (SELECT *
              FROM `{dataset_name}.{staging_table_name}` WHERE _sha IN UNNEST(@sha_list)
              UNION ALL
                SELECT * FROM `{dataset_name}.{metadata_table_name}` WHERE _sha IN UNNEST(@sha_list)) as metadata_staging_combine
              INNER JOIN (
                SELECT  _sha as _max_sha, MAX(_ts) as _max_ts FROM `{dataset_name}.{staging_table_name}` WHERE _sha IN UNNEST(@sha_list) GROUP BY _sha
                UNION ALL
                  SELECT  _sha as _max_sha, MAX(_ts) as _max_ts FROM `{dataset_name}.{metadata_table_name}` WHERE _sha IN UNNEST(@sha_list) GROUP BY _sha
              ) _max_metadata
            ON
              metadata_staging_combine._sha = _max_sha
            WHERE
              metadata_staging_combine._ts = _max_ts;
            """.format(
                dataset_name=bq_dataset.name,
                staging_table_name=self._get_table_name(bq_client.table_prefix, self.__get_staging_table_name()),
                metadata_table_name=self._get_table_name(bq_client.table_prefix, self.METADATA_TABLE_NAME))

            query_parameters = (
                bigquery.ArrayQueryParameter('sha_list', 'STRING', sha_list),
            )

            return self._query_sync(query, query_parameters, process_row=self.build_dict)

    def _query(self, sql_vars, select_fields, where):
        bq_client = self._connection

        with bq_client.get_cursor() as bq_dataset:
            limit = sql_vars.get('limit')
            limit = 'limit %s' % limit if limit is not None else ''

            query_sql = """
                 #Legacy SQL
                 select {select}
                  from {dataset_name}.{metadata_table_name}
                 where _ts IN (
                   SELECT MAX(_ts)
                   FROM {dataset_name}.{metadata_table_name}
                   GROUP BY _sha
                 ) AND ({where})
                 ORDER BY _sha
                 {limit}
             """.format(
                dataset_name=bq_dataset.name,
                metadata_table_name=self._get_table_name(bq_client.table_prefix, self.METADATA_TABLE_NAME),
                where=where or 'True',
                select=','.join(select_fields),
                limit=limit)

            query_sql = self.fill_in_vars(query_sql, sql_vars)

            data = self._query_sync(query_sql, use_legacy_sql=True, process_row=self.build_dict)

            for result in data:
                for key, val in result.items():
                    if isinstance(val, six.string_types):
                        result[key] = val.encode('utf8')

                yield unflatten(result, separator='.')

    def get_all_data(self, sha):
        pass

    def get_data_for_commit(self, sha, commit_sha):
        pass



# coding=utf-8

from click import option


def project_id_option(required=False):
    def decorator(f):
        return option('--projectId', '-p', type=int, metavar='<int>', required=required,
                      help='The project ID to find experiments. Use `mali projects list` to find your project IDs')(f)
    return decorator


def experiment_id_option(required=False):
    def decorator(f):
        return option('--experimentId', '-e', type=int, metavar='<int>', required=required,
                      help='The experiment ID. Use `mali experiments list` to find your experiment IDs')(f)
    return decorator


def metrics_option(required=False):
    def decorator(f):
        return option('--metrics', '-m', metavar='<json_string>', required=required,
                      help='The metrics of the experiment as a jsonified string. The key should be the metric '
                           'name with "ex" prefix e.g. "ex_cost". The value is the metric value in String, Float, '
                           'Integer or Boolean.')(f)
    return decorator


def model_weights_hash_option(required=False):
    def decorator(f):
        return option('--weightsHash', '-wh', metavar='<sha1_hex>', required=required,
                      help="The hexadecimal sha1 hash of the model's weights")(f)

    return decorator



# coding=utf-8
from ..connection_mixin import ConnectionMixin
from ..dulwich.refs import RefsContainer

SYMREF = b'ref: '


class MySqlRefsContainer(ConnectionMixin, RefsContainer):
    create_refs_sql = """
      CREATE TABLE
      IF NOT EXISTS refs (
        name VARCHAR(256) PRIMARY KEY,
        data TEXT
      );
    """

    def __init__(self, connection):
        super(MySqlRefsContainer, self).__init__(connection)
        self._create_table_if_needed()

    def get_packed_refs(self):
        return {}

    def read_loose_ref(self, name):
        """Read a reference and return its contents.

        If the reference file a symbolic reference, only read the first line of
        the file. Otherwise, only read the first 40 bytes.

        :param name: the refname to read, relative to refpath
        :return: The contents of the ref file, or None if the file does not
            exist.
        """

        with self._connection.get_cursor() as c:
            c.execute('SELECT data FROM refs WHERE name = %s LIMIT 1', (name,))
            data = c.fetchall()

        if len(data) == 0:
            return None

        data = data[0][0]

        header = data[:len(SYMREF)]
        if header == SYMREF:
            # Read only the first line
            return data.splitlines()[0]

        # Read only the first 40 bytes
        return data[:40].encode('ascii')

    def set_if_equals(self, name, old_ref, new_ref):
        """Set a refname to new_ref only if it currently equals old_ref.

        This method follows all symbolic references, and can be used to perform
        an atomic compare-and-swap operation.

        :param name: The refname to set.
        :param old_ref: The old sha the refname must refer to, or None to set
            unconditionally.
        :param new_ref: The new sha the refname will refer to.
        :return: True if the set was successful, False otherwise.
        """

        self._check_refname(name)
        try:
            real_names, _ = self.follow(name)
            real_name = real_names[-1]
        except (KeyError, IndexError):
            real_name = name

        real_name = real_name.decode('ascii')
        old_ref = old_ref.decode('ascii')
        new_ref = new_ref.decode('ascii')

        update_sql = """
            UPDATE refs SET data = %(new_data)s WHERE name = %(name)s AND data = %(old_data)s
        """

        if old_ref is not None:
            with self._connection.get_cursor() as c:
                c.execute(update_sql, dict(name=real_name, new_data=new_ref, old_data=old_ref))
                self._connection.commit()

        return True

    def add_if_new(self, name, ref):
        update_sql = """
            INSERT INTO refs (name, data)
            SELECT * FROM (SELECT %(name)s, %(data)s) AS tmp
            WHERE NOT EXISTS (
                SELECT name FROM refs WHERE name = %(name)s
            ) LIMIT 1;
        """
        self._check_refname(name)

        name = name.decode('ascii')
        ref = ref.decode('ascii')

        with self._connection.get_cursor() as c:
            c.execute(update_sql, dict(name=name, data=ref))
            self._connection.commit()

        return True

    def remove_if_equals(self, name, old_ref):
        raise NotImplementedError()

    def _create_connection(self, **kwargs):
        import mysql.connector

        user = kwargs.get('user')
        password = kwargs.get('password')
        host = kwargs.get('host')
        database = kwargs.get('database')

        return mysql.connector.connect(user=user, password=password, host=host, database=database)

    def _create_table_if_needed(self):
        with self._connection.get_cursor() as c:
            c.execute(self.create_refs_sql)
            self._connection.commit()



# coding=utf-8
from ..datastore_mixin import DatastoreMixin
from ..dulwich.refs import RefsContainer


SYMREF = b'ref: '


class DatastoreRefContainer(DatastoreMixin, RefsContainer):
    def __init__(self, connection):
        super(DatastoreRefContainer, self).__init__(connection)

    @property
    def _entity_kind(self):
        return 'DataVolumeRef'

    def get_packed_refs(self):
        return {}

    def set_if_equals(self, name, old_ref, new_ref):
        self._check_refname(name)

        name = name.decode('ascii')
        old_ref = old_ref.decode('ascii')
        new_ref = new_ref.decode('ascii')

        def txn(key):
            with datastore_client.transaction():
                ref_entity = datastore_client.get(key)

                if ref_entity is None:
                    return False

                if ref_entity['data'] != old_ref:
                    return False

                entity = self._connection.create_entity(key)
                entity['data'] = new_ref
                datastore_client.put(entity)

                return True

        with self._connection.get_cursor() as datastore_client:
            ref_key = self._build_key(datastore_client, name)

            return txn(ref_key)

    def add_if_new(self, name, ref):
        self._check_refname(name)

        name = name.decode('ascii')
        ref = ref.decode('ascii')

        def txn(key):
            with datastore_client.transaction():
                ref_entity = datastore_client.get(key)

                if ref_entity is not None:
                    return False

                entity = self._connection.create_entity(key)
                entity['data'] = ref
                datastore_client.put(entity)

                return True

        with self._connection.get_cursor() as datastore_client:
            ref_key = self._build_key(datastore_client, name)

            return txn(ref_key)

    def read_loose_ref(self, name):
        """Read a reference and return its contents.

        If the reference file a symbolic reference, only read the first line of
        the file. Otherwise, only read the first 40 bytes.

        :param name: the refname to read, relative to refpath
        :return: The contents of the ref file, or None if the file does not
            exist.
        """

        with self._connection.get_cursor() as datastore:
            ref_key = self._build_key(datastore, name.decode('utf8'))
            ref = datastore.get(ref_key)

            if ref is None:
                return None

            data = ref['data']

            header = data[:len(SYMREF)]
            if header == SYMREF:
                # Read only the first line
                return data.splitlines()[0]

            # Read only the first 40 bytes
            return data[:40].encode('ascii')

    def set_symbolic_ref(self, name, other):
        raise NotImplementedError(self.set_symbolic_ref)

    def allkeys(self):
        raise NotImplementedError(self.allkeys)

    def remove_if_equals(self, name, old_ref):
        raise NotImplementedError(self.remove_if_equals)



# coding=utf-8
from contextlib import closing
from .base_db_index import BaseMLIndex
from .cache import SqliteTreeCache


class SqliteMLIndex(BaseMLIndex):
    create_index_sql = """
      CREATE TABLE
      IF NOT EXISTS files (
        name text PRIMARY KEY,
        sha text not null,
        ctime integer not null,
        mtime integer not null,
        mode integer not null,
        uid integer not null,
        gid integer not null,
        size integer not null
      );
    """

    def __init__(self, connection, **kwargs):
        self.__tree_cache = SqliteTreeCache(connection)
        super(SqliteMLIndex, self).__init__(connection)

    def tree_cache(self):
        return self.__tree_cache

    def _get_hash_by_name_query(self):
        return "SELECT sha from files where name=? limit 1"

    def remove_path(self, name):
        with closing(self._conn.cursor()) as c:
            c.execute("DELETE from files where name=? limit 1", (name.decode('utf8'), ))
            self._conn.commit()

    def set_entries(self, entries):
        values = self._decode_entries(entries)

        with self._connection.get_cursor() as c:
            c.executemany("INSERT OR REPLACE INTO files VALUES (?, ?, ?, ?, ?, ?, ?, ?)", values)

            self.__tree_cache.set_entries(entries)

            self._connection.commit()

    def _create_table_if_needed(self):
        with self._connection.get_cursor() as c:
            self._connection.execute_multi(c, self.create_index_sql)
            self._connection.commit()

    def write(self):
        pass

    def iterblobs(self):
        with self._connection.get_cursor() as c:
            c.execute('SELECT name, sha, mode FROM files order by name')
            for row in c:
                yield row['name'].encode('utf8'), row['sha'].encode('utf8'), row['mode']

    def commit(self, object_store):
        from dulwich.index import commit_tree

        return commit_tree(object_store, self.iterblobs())

    def get_commit_id(self):
        return self.__tree_cache.get_commit_id()



# coding=utf-8
import os


def get_dist(package):
    from pkg_resources import get_distribution, DistributionNotFound

    dist = get_distribution(package)

    if os.environ.get('ML_SKIP_DIST_CHECK'):
        return dist

    # Normalize case for Windows systems
    dist_loc = os.path.normcase(dist.location)
    here = os.path.normcase(__file__)

    if not here.startswith(os.path.join(dist_loc, package)):
        # not installed, but there is another version that *is*
        raise DistributionNotFound

    return dist


def get_keywords(package):
    from pkg_resources import DistributionNotFound

    try:
        dist = get_dist(package)
    except DistributionNotFound:
        return None

    parsed_pkg_info = getattr(dist, '_parsed_pkg_info', None)

    if parsed_pkg_info is None:
        return None

    return parsed_pkg_info.get('Keywords')


def get_version(package):
    from pkg_resources import DistributionNotFound

    try:
        dist = get_dist(package)
    except DistributionNotFound:
        return None

    return str(dist.version)
